from datetime import timedelta

from django.utils import timezone

from scheduler.models import ScheduledTask
from scheduler.tests.testtools import SchedulerBaseCase, task_factory
from scheduler.tools import get_scheduled_task


class TestInternals(SchedulerBaseCase):
    def test_get_scheduled_job(self):
        task = task_factory(ScheduledTask, scheduled_time=timezone.now() - timedelta(hours=1))
        self.assertEqual(task, get_scheduled_task(task.TASK_TYPE, task.id))
        with self.assertRaises(ValueError):
            get_scheduled_task(task.TASK_TYPE, task.id + 1)
        with self.assertRaises(ValueError):
            get_scheduled_task('UNKNOWN_JOBTYPE', task.id)



import sys
from typing import Dict, Any

import click
from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand
from django.utils import timezone

from scheduler.models import TaskArg, TaskKwarg
from scheduler.tools import MODEL_NAMES


def job_model_str(model_str: str) -> str:
    if model_str.find('Job') == len(model_str) - 3:
        return model_str[:-3] + 'Task'
    return model_str


def create_job_from_dict(job_dict: Dict[str, Any], update):
    model = apps.get_model(app_label='scheduler', model_name=job_model_str(job_dict['model']))
    existing_job = model.objects.filter(name=job_dict['name']).first()
    if existing_job:
        if update:
            click.echo(f'Found existing job "{existing_job}, removing it to be reinserted"')
            existing_job.delete()
        else:
            click.echo(f'Found existing job "{existing_job}", skipping')
            return
    kwargs = dict(job_dict)
    del kwargs['model']
    del kwargs['callable_args']
    del kwargs['callable_kwargs']
    if kwargs.get('scheduled_time', None):
        target = timezone.datetime.fromisoformat(kwargs['scheduled_time'])
        if not settings.USE_TZ and not timezone.is_naive(target):
            target = timezone.make_naive(target)
        kwargs['scheduled_time'] = target
    model_fields = set(map(lambda field: field.attname, model._meta.get_fields()))
    keys_to_ignore = list(filter(lambda k: k not in model_fields, kwargs.keys()))
    for k in keys_to_ignore:
        del kwargs[k]
    scheduled_job = model.objects.create(**kwargs)
    click.echo(f'Created job {scheduled_job}')
    content_type = ContentType.objects.get_for_model(scheduled_job)

    for arg in job_dict['callable_args']:
        TaskArg.objects.create(
            content_type=content_type, object_id=scheduled_job.id, **arg, )
    for arg in job_dict['callable_kwargs']:
        TaskKwarg.objects.create(
            content_type=content_type, object_id=scheduled_job.id, **arg, )


class Command(BaseCommand):
    """
    Import scheduled jobs
    """
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            '-f', '--format',
            action='store',
            choices=['json', 'yaml'],
            default='json',
            dest='format',
            help='format of input',
        )
        parser.add_argument(
            '--filename',
            action='store',
            dest='filename',
            help='File name to load (otherwise loads from standard input)',
        )

        parser.add_argument(
            '-r', '--reset',
            action='store_true',
            dest='reset',
            help='Remove all currently scheduled jobs before importing',
        )
        parser.add_argument(
            '-u', '--update',
            action='store_true',
            dest='update',
            help='Update existing records',
        )

    def handle(self, *args, **options):
        file = open(options.get('filename')) if options.get("filename") else sys.stdin
        jobs = list()
        if options.get("format") == 'json':
            import json
            try:
                jobs = json.load(file)
            except json.decoder.JSONDecodeError:
                click.echo('Error decoding json', err=True)
                exit(1)
        elif options.get("format") == 'yaml':
            try:
                import yaml
            except ImportError:
                click.echo("Aborting. LibYAML is not installed.")
                exit(1)
            # Disable YAML alias
            yaml.Dumper.ignore_aliases = lambda *x: True
            jobs = yaml.load(file, yaml.SafeLoader)

        if options.get('reset'):
            for model_name in MODEL_NAMES:
                model = apps.get_model(app_label='scheduler', model_name=model_name)
                model.objects.all().delete()

        for job in jobs:
            create_job_from_dict(job, update=options.get('update'))



import time

import click
from django.core.management.base import BaseCommand

from scheduler.views import get_statistics

ANSI_LIGHT_GREEN = "\033[1;32m"
ANSI_LIGHT_WHITE = "\033[1;37m"
ANSI_RESET = "\033[0m"

KEYS = ('jobs', 'started_jobs', 'deferred_jobs', 'finished_jobs', 'canceled_jobs', 'workers')


class Command(BaseCommand):
    """
    Print statistics
    """
    help = __doc__

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.table_width = 80
        self.interval = None

    def add_arguments(self, parser):
        parser.add_argument(
            '-j', '--json', action='store_true', dest='json',
            help='Output statistics as JSON', )

        parser.add_argument(
            '-y', '--yaml', action='store_true', dest='yaml',
            help='Output statistics as YAML',
        )

        parser.add_argument(
            '-i', '--interval', dest='interval', type=float,
            help='Poll statistics every N seconds',
        )

    def _print_separator(self):
        click.echo('-' * self.table_width)

    def _print_stats_dashboard(self, statistics, prev_stats=None):
        if self.interval:
            click.clear()
        click.echo()
        click.echo("Django-Scheduler CLI Dashboard")
        click.echo()
        self._print_separator()
        click.echo(f'| {"Name":<16} |    Queued |    Active |  Deferred |'
                   f'  Finished |'
                   f'  Canceled |'
                   f'   Workers |')
        self._print_separator()
        for ind, queue in enumerate(statistics["queues"]):
            vals = list((queue[k] for k in KEYS))
            # Deal with colors
            if prev_stats and len(prev_stats['queues']) > ind:
                prev = prev_stats["queues"][ind]
                prev_vals = (prev[k] for k in KEYS)
                colors = [ANSI_LIGHT_GREEN
                          if vals[i] != prev_vals[i] else ANSI_LIGHT_WHITE
                          for i in range(len(prev_vals))
                          ]
            else:
                colors = [ANSI_LIGHT_WHITE for _ in range(len(vals))]
            to_print = ' | '.join([f'{colors[i]}{vals[i]:9}{ANSI_RESET}' for i in range(len(vals))])
            click.echo(f'| {queue["name"]:<16} | {to_print} |', color=True)

        self._print_separator()

        if self.interval:
            click.echo()
            click.echo("Press 'Ctrl+c' to quit")

    def handle(self, *args, **options):

        if options.get("json"):
            import json
            click.secho(json.dumps(get_statistics(), indent=2), )
            return

        if options.get("yaml"):
            try:
                import yaml
            except ImportError:
                click.secho("Aborting. yaml not supported", err=True, fg='red')
                return

            click.secho(yaml.dump(get_statistics(), default_flow_style=False), )
            return

        self.interval = options.get("interval")

        if not self.interval or self.interval < 0:
            self._print_stats_dashboard(get_statistics())
            return

        try:
            prev = None
            while True:
                statistics = get_statistics()
                self._print_stats_dashboard(statistics, prev)
                prev = statistics
                time.sleep(self.interval)
        except KeyboardInterrupt:
            pass



from datetime import datetime
from typing import Callable

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from scheduler import tools

ARG_TYPE_TYPES_DICT = {
    'str': str,
    'int': int,
    'bool': bool,
    'datetime': datetime,
    'callable': Callable,
}


class BaseTaskArg(models.Model):
    class ArgType(models.TextChoices):
        STR = 'str', _('string')
        INT = 'int', _('int')
        BOOL = 'bool', _('boolean')
        DATETIME = 'datetime', _('datetime')
        CALLABLE = 'callable', _('callable')

    arg_type = models.CharField(
        _('Argument Type'), max_length=12, choices=ArgType.choices, default=ArgType.STR,
    )
    val = models.CharField(_('Argument Value'), blank=True, max_length=255)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey()

    def clean(self):
        if self.arg_type not in ARG_TYPE_TYPES_DICT:
            raise ValidationError({
                'arg_type': ValidationError(
                    _(f'Could not parse {self.arg_type}, options are: {ARG_TYPE_TYPES_DICT.keys()}'), code='invalid')
            })
        try:
            if self.arg_type == 'callable':
                tools.callable_func(self.val)
            elif self.arg_type == 'datetime':
                datetime.fromisoformat(self.val)
            elif self.arg_type == 'bool':
                if self.val.lower() not in {'true', 'false'}:
                    raise ValidationError
            elif self.arg_type == 'int':
                int(self.val)
        except Exception:
            raise ValidationError({
                'arg_type': ValidationError(
                    _(f'Could not parse {self.val} as {self.arg_type}'), code='invalid')
            })

    def save(self, **kwargs):
        super(BaseTaskArg, self).save(**kwargs)
        self.content_object.save()

    def delete(self, **kwargs):
        super(BaseTaskArg, self).delete(**kwargs)
        self.content_object.save()

    def value(self):
        if self.arg_type == 'callable':
            res = tools.callable_func(self.val)()
        elif self.arg_type == 'datetime':
            res = datetime.fromisoformat(self.val)
        elif self.arg_type == 'bool':
            res = self.val.lower() == 'true'
        else:
            res = ARG_TYPE_TYPES_DICT[self.arg_type](self.val)
        return res

    class Meta:
        abstract = True
        ordering = ['id']


class TaskArg(BaseTaskArg):
    def __str__(self):
        return f'TaskArg[arg_type={self.arg_type},value={self.value()}]'


class TaskKwarg(BaseTaskArg):
    key = models.CharField(max_length=255)

    def __str__(self):
        key, value = self.value()
        return f'TaskKwarg[key={key},arg_type={self.arg_type},value={self.val}]'

    def value(self):
        return self.key, super(TaskKwarg, self).value()



from datetime import timedelta

from django.core.exceptions import ValidationError
from django.test import override_settings
from django.utils import timezone

from scheduler import settings
from scheduler.models import RepeatableTask
from scheduler.tests.test_models import BaseTestCases
from .testtools import (
    task_factory, _get_job_from_scheduled_registry)


class TestRepeatableTask(BaseTestCases.TestSchedulableJob):
    TaskModelClass = RepeatableTask

    def test_unschedulable_old_job(self):
        job = task_factory(self.TaskModelClass, scheduled_time=timezone.now() - timedelta(hours=1), repeat=0)
        self.assertFalse(job.is_scheduled())

    def test_schedulable_old_job_repeat_none(self):
        # If repeat is None, the job should be scheduled
        job = task_factory(self.TaskModelClass, scheduled_time=timezone.now() - timedelta(hours=1), repeat=None)
        self.assertTrue(job.is_scheduled())

    def test_clean(self):
        job = task_factory(self.TaskModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 1
        job.result_ttl = -1
        self.assertIsNone(job.clean())

    def test_clean_seconds(self):
        job = task_factory(self.TaskModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 60
        job.result_ttl = -1
        job.interval_unit = 'seconds'
        self.assertIsNone(job.clean())

    @override_settings(SCHEDULER_CONFIG={
        'SCHEDULER_INTERVAL': 10,
    })
    def test_clean_too_frequent(self):
        job = task_factory(self.TaskModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 2  # Smaller than 10
        job.result_ttl = -1
        job.interval_unit = 'seconds'
        with self.assertRaises(ValidationError):
            job.clean_interval_unit()

    def test_clean_not_multiple(self):
        job = task_factory(self.TaskModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 121
        job.interval_unit = 'seconds'
        with self.assertRaises(ValidationError):
            job.clean_interval_unit()

    def test_clean_short_result_ttl(self):
        job = task_factory(self.TaskModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 1
        job.repeat = 1
        job.result_ttl = 3599
        job.interval_unit = 'hours'
        job.repeat = 42
        with self.assertRaises(ValidationError):
            job.clean_result_ttl()

    def test_clean_indefinite_result_ttl(self):
        job = task_factory(self.TaskModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 1
        job.result_ttl = -1
        job.interval_unit = 'hours'
        job.clean_result_ttl()

    def test_clean_undefined_result_ttl(self):
        job = task_factory(self.TaskModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 1
        job.interval_unit = 'hours'
        job.clean_result_ttl()

    def test_interval_seconds_weeks(self):
        job = task_factory(self.TaskModelClass, interval=2, interval_unit='weeks')
        self.assertEqual(1209600.0, job.interval_seconds())

    def test_interval_seconds_days(self):
        job = task_factory(self.TaskModelClass, interval=2, interval_unit='days')
        self.assertEqual(172800.0, job.interval_seconds())

    def test_interval_seconds_hours(self):
        job = task_factory(self.TaskModelClass, interval=2, interval_unit='hours')
        self.assertEqual(7200.0, job.interval_seconds())

    def test_interval_seconds_minutes(self):
        job = task_factory(self.TaskModelClass, interval=15, interval_unit='minutes')
        self.assertEqual(900.0, job.interval_seconds())

    def test_interval_seconds_seconds(self):
        job = RepeatableTask(interval=15, interval_unit='seconds')
        self.assertEqual(15.0, job.interval_seconds())

    def test_interval_display(self):
        job = task_factory(self.TaskModelClass, interval=15, interval_unit='minutes')
        self.assertEqual(job.interval_display(), '15 minutes')

    def test_result_interval(self):
        job = task_factory(self.TaskModelClass, )
        entry = _get_job_from_scheduled_registry(job)
        self.assertEqual(entry.meta['interval'], 3600)

    def test_repeat(self):
        job = task_factory(self.TaskModelClass, repeat=10)
        entry = _get_job_from_scheduled_registry(job)
        self.assertEqual(entry.meta['repeat'], 10)

    def test_repeat_old_job_exhausted(self):
        base_time = timezone.now()
        job = task_factory(self.TaskModelClass, scheduled_time=base_time - timedelta(hours=10), repeat=10)
        self.assertEqual(job.is_scheduled(), False)

    def test_repeat_old_job_last_iter(self):
        base_time = timezone.now()
        job = task_factory(self.TaskModelClass, scheduled_time=base_time - timedelta(hours=9, minutes=30), repeat=10)
        self.assertEqual(job.repeat, 0)
        self.assertEqual(job.is_scheduled(), True)

    def test_repeat_old_job_remaining(self):
        base_time = timezone.now()
        job = task_factory(self.TaskModelClass, scheduled_time=base_time - timedelta(minutes=30), repeat=5)
        self.assertEqual(job.repeat, 4)
        self.assertEqual(job.scheduled_time, base_time + timedelta(minutes=30))
        self.assertEqual(job.is_scheduled(), True)

    def test_repeat_none_interval_2_min(self):
        base_time = timezone.now()
        job = task_factory(self.TaskModelClass, scheduled_time=base_time - timedelta(minutes=29), repeat=None)
        job.interval = 120
        job.interval_unit = 'seconds'
        job.schedule()
        self.assertTrue(job.scheduled_time > base_time)
        self.assertTrue(job.is_scheduled())

    def test_check_rescheduled_after_execution(self):
        task = task_factory(self.TaskModelClass, scheduled_time=timezone.now() + timedelta(seconds=1), repeat=10)
        queue = task.rqueue
        first_run_id = task.job_id
        entry = queue.fetch_job(first_run_id)
        queue.run_sync(entry)
        task.refresh_from_db()
        self.assertEqual(task.failed_runs, 0)
        self.assertIsNone(task.last_failed_run)
        self.assertEqual(task.successful_runs, 1)
        self.assertIsNotNone(task.last_successful_run)
        self.assertTrue(task.is_scheduled())
        self.assertNotEqual(task.job_id, first_run_id)

    def test_check_rescheduled_after_execution_failed_job(self):
        task = task_factory(
            self.TaskModelClass, callable_name='scheduler.tests.jobs.failing_job',
            scheduled_time=timezone.now() + timedelta(seconds=1),
            repeat=10, )
        queue = task.rqueue
        first_run_id = task.job_id
        entry = queue.fetch_job(first_run_id)
        queue.run_sync(entry)
        task.refresh_from_db()
        self.assertEqual(task.failed_runs, 1)
        self.assertIsNotNone(task.last_failed_run)
        self.assertEqual(task.successful_runs, 0)
        self.assertIsNone(task.last_successful_run)
        self.assertTrue(task.is_scheduled())
        self.assertNotEqual(task.job_id, first_run_id)

    def test_check_not_rescheduled_after_last_repeat(self):
        task = task_factory(
            self.TaskModelClass,
            scheduled_time=timezone.now() + timedelta(seconds=1),
            repeat=1,
        )
        queue = task.rqueue
        first_run_id = task.job_id
        entry = queue.fetch_job(first_run_id)
        queue.run_sync(entry)
        task.refresh_from_db()
        self.assertEqual(task.failed_runs, 0)
        self.assertIsNone(task.last_failed_run)
        self.assertEqual(task.successful_runs, 1)
        self.assertIsNotNone(task.last_successful_run)
        self.assertNotEqual(task.job_id, first_run_id)



# Generated by Django 4.1.7 on 2023-03-16 20:59

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('scheduler', '0009_alter_jobarg_arg_type_alter_jobarg_val_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='Queue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'permissions': [['view', 'Access admin page']],
                'managed': False,
                'default_permissions': (),
            },
        ),
    ]



import redis
from django.contrib import admin, messages
from django.contrib.contenttypes.admin import GenericStackedInline
from django.utils.translation import gettext_lazy as _

from scheduler import tools
from scheduler.models import CronTask, TaskArg, TaskKwarg, RepeatableTask, ScheduledTask
from scheduler.settings import SCHEDULER_CONFIG, logger
from scheduler.tools import get_job_executions


class HiddenMixin(object):
    class Media:
        js = ['admin/js/jquery.init.js', ]


class JobArgInline(HiddenMixin, GenericStackedInline):
    model = TaskArg
    extra = 0
    fieldsets = (
        (None, {
            'fields': (('arg_type', 'val',),),
        }),
    )


class JobKwargInline(HiddenMixin, GenericStackedInline):
    model = TaskKwarg
    extra = 0
    fieldsets = (
        (None, {
            'fields': (('key',), ('arg_type', 'val',),),
        }),
    )


_LIST_DISPLAY_EXTRA = dict(
    CronTask=('cron_string', 'next_run', 'successful_runs', 'last_successful_run', 'failed_runs', 'last_failed_run',),
    ScheduledTask=('scheduled_time',),
    RepeatableTask=(
        'scheduled_time', 'interval_display', 'successful_runs', 'last_successful_run', 'failed_runs',
        'last_failed_run',),
)
_FIELDSET_EXTRA = dict(
    CronTask=(
        'cron_string', 'timeout', 'result_ttl',
        ('successful_runs', 'last_successful_run',),
        ('failed_runs', 'last_failed_run',),
    ),
    ScheduledTask=('scheduled_time', 'timeout', 'result_ttl'),
    RepeatableTask=(
        'scheduled_time',
        ('interval', 'interval_unit',),
        'repeat', 'timeout', 'result_ttl',
        ('successful_runs', 'last_successful_run',),
        ('failed_runs', 'last_failed_run',),
    ),
)


@admin.register(CronTask, ScheduledTask, RepeatableTask)
class TaskAdmin(admin.ModelAdmin):
    """TaskAdmin admin view for all task models.
    Using the _LIST_DISPLAY_EXTRA and _FIELDSET_EXTRA additional data for each model.
    """

    save_on_top = True
    change_form_template = 'admin/scheduler/change_form.html'
    actions = ['disable_selected', 'enable_selected', 'enqueue_job_now', ]
    inlines = [JobArgInline, JobKwargInline, ]
    list_filter = ('enabled',)
    list_display = ('enabled', 'name', 'job_id', 'function_string', 'is_scheduled', 'queue',)
    list_display_links = ('name',)
    readonly_fields = ('job_id',)
    fieldsets = (
        (None, {
            'fields': ('name', 'callable', 'enabled', 'at_front',),
        }),
        (_('RQ Settings'), {
            'fields': ('queue', 'job_id',),
        }),
    )

    def get_list_display(self, request):
        if self.model.__name__ not in _LIST_DISPLAY_EXTRA:
            raise ValueError(f'Unrecognized model {self.model}')
        return TaskAdmin.list_display + _LIST_DISPLAY_EXTRA[self.model.__name__]

    def get_fieldsets(self, request, obj=None):
        if self.model.__name__ not in _FIELDSET_EXTRA:
            raise ValueError(f'Unrecognized model {self.model}')
        return TaskAdmin.fieldsets + ((_('Scheduling'), {
            'fields': _FIELDSET_EXTRA[self.model.__name__],
        }),)

    @admin.display(description='Next run')
    def next_run(self, o: CronTask):
        return tools.get_next_cron_time(o.cron_string)

    def change_view(self, request, object_id, form_url='', extra_context=None):
        extra = extra_context or {}
        obj = self.get_object(request, object_id)
        try:
            execution_list = get_job_executions(obj.queue, obj)
        except redis.ConnectionError as e:
            logger.warn(f'Could not get job executions: {e}')
            execution_list = list()
        paginator = self.get_paginator(request, execution_list, SCHEDULER_CONFIG['EXECUTIONS_IN_PAGE'])
        page_number = request.GET.get('p', 1)
        page_obj = paginator.get_page(page_number)
        page_range = paginator.get_elided_page_range(page_obj.number)

        extra.update({
            "pagination_required": paginator.count > SCHEDULER_CONFIG['EXECUTIONS_IN_PAGE'],
            'executions': page_obj,
            'page_range': page_range,
            'page_var': 'p',
        })

        return super(TaskAdmin, self).change_view(
            request, object_id, form_url, extra_context=extra)

    def delete_queryset(self, request, queryset):
        for job in queryset:
            job.unschedule()
        super(TaskAdmin, self).delete_queryset(request, queryset)

    def delete_model(self, request, obj):
        obj.unschedule()
        super(TaskAdmin, self).delete_model(request, obj)

    @admin.action(description=_("Disable selected %(verbose_name_plural)s"), permissions=('change',))
    def disable_selected(self, request, queryset):
        rows_updated = 0
        for obj in queryset.filter(enabled=True).iterator():
            obj.enabled = False
            obj.unschedule()
            rows_updated += 1

        message_bit = "1 job was" if rows_updated == 1 else f"{rows_updated} jobs were"

        level = messages.WARNING if not rows_updated else messages.INFO
        self.message_user(request, f"{message_bit} successfully disabled and unscheduled.", level=level)

    @admin.action(description=_("Enable selected %(verbose_name_plural)s"), permissions=('change',))
    def enable_selected(self, request, queryset):
        rows_updated = 0
        for obj in queryset.filter(enabled=False).iterator():
            obj.enabled = True
            obj.save()
            rows_updated += 1

        message_bit = "1 job was" if rows_updated == 1 else f"{rows_updated} jobs were"
        level = messages.WARNING if not rows_updated else messages.INFO
        self.message_user(request, f"{message_bit} successfully enabled and scheduled.", level=level)

    @admin.action(description="Enqueue now", permissions=('change',))
    def enqueue_job_now(self, request, queryset):
        task_names = []
        for task in queryset:
            task.enqueue_to_run()
            task_names.append(task.name)
        self.message_user(request, f"The following jobs have been enqueued: {', '.join(task_names)}", )



# Generated by Django 4.2.5 on 2023-10-08 16:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('scheduler', '0015_rename_cronjob_crontask_and_more'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='JobArg',
            new_name='TaskArg',
        ),
        migrations.RenameModel(
            old_name='JobKwarg',
            new_name='TaskKwarg',
        ),
        migrations.AlterModelOptions(
            name='crontask',
            options={'ordering': ('name',), 'verbose_name': 'Cron Task', 'verbose_name_plural': 'Cron Tasks'},
        ),
        migrations.AlterModelOptions(
            name='repeatabletask',
            options={'ordering': ('name',), 'verbose_name': 'Repeatable Task', 'verbose_name_plural': 'Repeatable Tasks'},
        ),
        migrations.AlterModelOptions(
            name='scheduledtask',
            options={'ordering': ('name',), 'verbose_name': 'Scheduled Task', 'verbose_name_plural': 'Scheduled Tasks'},
        ),
    ]



from typing import List, Dict

import redis
from redis.sentinel import Sentinel

from .rq_classes import JobExecution, DjangoQueue, DjangoWorker
from .settings import get_config
from .settings import logger

_CONNECTION_PARAMS = {
    'URL',
    'DB',
    'USE_REDIS_CACHE',
    'UNIX_SOCKET_PATH',
    'HOST',
    'PORT',
    'PASSWORD',
    'SENTINELS',
    'MASTER_NAME',
    'SOCKET_TIMEOUT',
    'SSL',
    'CONNECTION_KWARGS',
}


class QueueNotFoundError(Exception):
    pass


def _get_redis_connection(config, use_strict_redis=False):
    """
    Returns a redis connection from a connection config
    """
    if get_config('FAKEREDIS'):
        import fakeredis
        redis_cls = fakeredis.FakeRedis if use_strict_redis else fakeredis.FakeStrictRedis
    else:
        redis_cls = redis.StrictRedis if use_strict_redis else redis.Redis
    logger.debug(f'Getting connection for {config}')
    if 'URL' in config:
        if config.get('SSL') or config.get('URL').startswith('rediss://'):
            return redis_cls.from_url(
                config['URL'],
                db=config.get('DB'),
                ssl_cert_reqs=config.get('SSL_CERT_REQS', 'required'),
            )
        else:
            return redis_cls.from_url(
                config['URL'],
                db=config.get('DB'),
            )
    if 'UNIX_SOCKET_PATH' in config:
        return redis_cls(unix_socket_path=config['UNIX_SOCKET_PATH'], db=config['DB'])

    if 'SENTINELS' in config:
        connection_kwargs = {
            'db': config.get('DB'),
            'password': config.get('PASSWORD'),
            'username': config.get('USERNAME'),
            'socket_timeout': config.get('SOCKET_TIMEOUT'),
        }
        connection_kwargs.update(config.get('CONNECTION_KWARGS', {}))
        sentinel_kwargs = config.get('SENTINEL_KWARGS', {})
        sentinel = Sentinel(config['SENTINELS'], sentinel_kwargs=sentinel_kwargs, **connection_kwargs)
        return sentinel.master_for(
            service_name=config['MASTER_NAME'],
            redis_class=redis_cls,
        )

    return redis_cls(
        host=config['HOST'],
        port=config['PORT'],
        db=config.get('DB', 0),
        username=config.get('USERNAME', None),
        password=config.get('PASSWORD'),
        ssl=config.get('SSL', False),
        ssl_cert_reqs=config.get('SSL_CERT_REQS', 'required'),
        **config.get('REDIS_CLIENT_KWARGS', {})
    )


def get_connection(queue_settings, use_strict_redis=False):
    """Returns a Redis connection to use based on parameters in SCHEDULER_QUEUES
    """
    return _get_redis_connection(queue_settings, use_strict_redis)


def get_queue(
        name='default',
        default_timeout=None, is_async=None,
        autocommit=None,
        connection=None,
        **kwargs
) -> DjangoQueue:
    """Returns an DjangoQueue using parameters defined in `SCHEDULER_QUEUES`
    """
    from .settings import QUEUES
    if name not in QUEUES:
        raise QueueNotFoundError(f'Queue {name} not found, queues={QUEUES.keys()}')
    queue_settings = QUEUES[name]
    if is_async is None:
        is_async = queue_settings.get('ASYNC', True)

    if default_timeout is None:
        default_timeout = queue_settings.get('DEFAULT_TIMEOUT')
    if connection is None:
        connection = get_connection(queue_settings)
    return DjangoQueue(
        name,
        default_timeout=default_timeout,
        connection=connection,
        is_async=is_async,
        autocommit=autocommit,
        **kwargs
    )


def get_all_workers():
    from .settings import QUEUES
    workers = set()
    for queue_name in QUEUES:
        connection = get_connection(QUEUES[queue_name])
        try:
            curr_workers = set(DjangoWorker.all(connection=connection))
            workers.update(curr_workers)
        except redis.ConnectionError as e:
            logger.error(f'Could not connect for queue {queue_name}: {e}')
    return workers


def _queues_share_connection_params(q1_params: Dict, q2_params: Dict):
    """Check that both queues share the same connection parameters
    """
    return all(
        ((p not in q1_params and p not in q2_params)
         or (q1_params.get(p, None) == q2_params.get(p, None)))
        for p in _CONNECTION_PARAMS)


def get_queues(*queue_names, **kwargs) -> List[DjangoQueue]:
    """Return queue instances from specified queue names.
    All instances must use the same Redis connection.
    """
    from .settings import QUEUES

    kwargs['job_class'] = JobExecution
    queue_params = QUEUES[queue_names[0]]
    queues = [get_queue(queue_names[0], **kwargs)]
    # perform consistency checks while building return list
    for name in queue_names[1:]:
        if not _queues_share_connection_params(queue_params, QUEUES[name]):
            raise ValueError(
                f'Queues must have the same redis connection. "{name}" and'
                f' "{queue_names[0]}" have different connections')
        queue = get_queue(name, **kwargs)
        queues.append(queue)

    return queues



# Generated by Django 3.2.12 on 2022-03-29 21:07

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('scheduler', '0002_alter_cronjob_id_alter_repeatablejob_id_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cronjob',
            name='cron_string',
            field=models.CharField(help_text='Define the schedule in a crontab like syntax. Times are in UTC.',
                                   max_length=64, verbose_name='cron string'),
        ),
        migrations.AlterField(
            model_name='cronjob',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='interval_unit',
            field=models.CharField(
                choices=[('seconds', 'seconds'), ('minutes', 'minutes'), ('hours', 'hours'), ('days', 'days'),
                         ('weeks', 'weeks')], default='hours', max_length=12, verbose_name='interval unit'),
        ),
        migrations.AlterField(
            model_name='scheduledjob',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.CreateModel(
            name='JobKwarg',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('arg_type', models.CharField(
                    choices=[('str_val', 'string'), ('int_val', 'int'), ('bool_val', 'boolean'),
                             ('datetime_val', 'Datetime')], default='str_val', max_length=12,
                    verbose_name='Argument Type')),
                ('str_val', models.CharField(blank=True, max_length=255, verbose_name='String Value')),
                ('int_val', models.IntegerField(blank=True, null=True, verbose_name='Int Value')),
                ('bool_val', models.BooleanField(default=False, verbose_name='Boolean Value')),
                ('datetime_val', models.DateTimeField(blank=True, null=True, verbose_name='Datetime Value')),
                ('object_id', models.PositiveIntegerField()),
                ('key', models.CharField(max_length=255)),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
            ],
            options={
                'ordering': ['id'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='JobArg',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('arg_type', models.CharField(
                    choices=[('str_val', 'string'), ('int_val', 'int'), ('bool_val', 'boolean'),
                             ('datetime_val', 'Datetime')], default='str_val', max_length=12,
                    verbose_name='Argument Type')),
                ('str_val', models.CharField(blank=True, max_length=255, verbose_name='String Value')),
                ('int_val', models.IntegerField(blank=True, null=True, verbose_name='Int Value')),
                ('bool_val', models.BooleanField(default=False, verbose_name='Boolean Value')),
                ('datetime_val', models.DateTimeField(blank=True, null=True, verbose_name='Datetime Value')),
                ('object_id', models.PositiveIntegerField()),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
            ],
            options={
                'ordering': ['id'],
                'abstract': False,
            },
        ),
    ]



import json
import os
import tempfile
from unittest import mock

import yaml
from django.core.management import call_command
from django.test import TestCase

from scheduler.models import ScheduledTask, RepeatableTask
from scheduler.queues import get_queue
from scheduler.tests.jobs import failing_job, test_job
from scheduler.tests.testtools import task_factory
from . import test_settings  # noqa
from .test_views import BaseTestCase
from ..tools import create_worker


class RqworkerTestCase(TestCase):

    def test_rqworker__no_queues_params(self):
        queue = get_queue('default')

        # enqueue some jobs that will fail
        jobs = []
        job_ids = []
        for _ in range(0, 3):
            job = queue.enqueue(failing_job)
            jobs.append(job)
            job_ids.append(job.id)

        # Create a worker to execute these jobs
        call_command('rqworker', fork_job_execution=False, burst=True)

        # check if all jobs are really failed
        for job in jobs:
            self.assertTrue(job.is_failed)

    def test_rqworker__run_jobs(self):
        queue = get_queue('default')

        # enqueue some jobs that will fail
        jobs = []
        job_ids = []
        for _ in range(0, 3):
            job = queue.enqueue(failing_job)
            jobs.append(job)
            job_ids.append(job.id)

        # Create a worker to execute these jobs
        call_command('rqworker', 'default', fork_job_execution=False, burst=True)

        # check if all jobs are really failed
        for job in jobs:
            self.assertTrue(job.is_failed)

    def test_rqworker__worker_with_two_queues(self):
        queue = get_queue('default')
        queue2 = get_queue('django_tasks_scheduler_test')

        # enqueue some jobs that will fail
        jobs = []
        job_ids = []
        for _ in range(0, 3):
            job = queue.enqueue(failing_job)
            jobs.append(job)
            job_ids.append(job.id)
        job = queue2.enqueue(failing_job)
        jobs.append(job)
        job_ids.append(job.id)

        # Create a worker to execute these jobs
        call_command('rqworker', 'default', 'django_tasks_scheduler_test', fork_job_execution=False, burst=True)

        # check if all jobs are really failed
        for job in jobs:
            self.assertTrue(job.is_failed)

    def test_rqworker__worker_with_one_queue__does_not_perform_other_queue_job(self):
        queue = get_queue('default')
        queue2 = get_queue('django_tasks_scheduler_test')

        job = queue.enqueue(failing_job)
        other_job = queue2.enqueue(failing_job)

        # Create a worker to execute these jobs
        call_command('rqworker', 'default', fork_job_execution=False, burst=True)
        # assert
        self.assertTrue(job.is_failed)
        self.assertTrue(other_job.is_queued)


class RqstatsTest(TestCase):
    def test_rqstats__does_not_fail(self):
        call_command('rqstats', '-j')
        call_command('rqstats', '-y')
        call_command('rqstats')


class DeleteFailedExecutionsTest(BaseTestCase):
    def test_delete_failed_executions__delete_jobs(self):
        queue = get_queue('default')
        call_command('delete_failed_executions', queue='default')
        queue.enqueue(failing_job)
        worker = create_worker('default')
        worker.work(burst=True)
        self.assertEqual(1, len(queue.failed_job_registry))
        call_command('delete_failed_executions', queue='default')
        self.assertEqual(0, len(queue.failed_job_registry))


class RunJobTest(TestCase):
    def test_run_job__should_schedule_job(self):
        queue = get_queue('default')
        queue.empty()
        func_name = f'{test_job.__module__}.{test_job.__name__}'
        # act
        call_command('run_job', func_name, queue='default')
        # assert
        job_list = queue.get_jobs()
        self.assertEqual(1, len(job_list))
        self.assertEqual(func_name + '()', job_list[0].get_call_string())


class ExportTest(TestCase):
    def setUp(self) -> None:
        self.tmpfile = tempfile.NamedTemporaryFile()

    def tearDown(self) -> None:
        os.remove(self.tmpfile.name)

    def test_export__should_export_job(self):
        jobs = list()
        jobs.append(task_factory(ScheduledTask, enabled=True))
        jobs.append(task_factory(RepeatableTask, enabled=True))

        # act
        call_command('export', filename=self.tmpfile.name)
        # assert
        result = json.load(self.tmpfile)
        self.assertEqual(len(jobs), len(result))
        self.assertEqual(result[0], jobs[0].to_dict())
        self.assertEqual(result[1], jobs[1].to_dict())

    def test_export__should_export_enabled_jobs_only(self):
        jobs = list()
        jobs.append(task_factory(ScheduledTask, enabled=True))
        jobs.append(task_factory(RepeatableTask, enabled=False))

        # act
        call_command('export', filename=self.tmpfile.name, enabled=True)
        # assert
        result = json.load(self.tmpfile)
        self.assertEqual(len(jobs) - 1, len(result))
        self.assertEqual(result[0], jobs[0].to_dict())

    def test_export__should_export_job_yaml_without_yaml_lib(self):
        jobs = list()
        jobs.append(task_factory(ScheduledTask, enabled=True))
        jobs.append(task_factory(RepeatableTask, enabled=True))

        # act
        with mock.patch.dict('sys.modules', {'yaml': None}):
            with self.assertRaises(SystemExit) as cm:
                call_command('export', filename=self.tmpfile.name, format='yaml')
            self.assertEqual(cm.exception.code, 1)

    def test_export__should_export_job_yaml_green(self):
        jobs = list()
        jobs.append(task_factory(ScheduledTask, enabled=True))
        jobs.append(task_factory(RepeatableTask, enabled=True))

        # act
        call_command('export', filename=self.tmpfile.name, format='yaml')
        # assert
        result = yaml.load(self.tmpfile, yaml.SafeLoader)
        self.assertEqual(len(jobs), len(result))
        self.assertEqual(result[0], jobs[0].to_dict())
        self.assertEqual(result[1], jobs[1].to_dict())


class ImportTest(TestCase):
    def setUp(self) -> None:
        self.tmpfile = tempfile.NamedTemporaryFile(mode='w')

    def tearDown(self) -> None:
        os.remove(self.tmpfile.name)

    def test_import__should_schedule_job(self):
        jobs = list()
        jobs.append(task_factory(ScheduledTask, enabled=True, instance_only=True))
        jobs.append(task_factory(RepeatableTask, enabled=True, instance_only=True))
        res = json.dumps([j.to_dict() for j in jobs])
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        call_command('import', filename=self.tmpfile.name)
        # assert
        self.assertEqual(1, ScheduledTask.objects.count())
        db_job = ScheduledTask.objects.first()
        attrs = ['name', 'queue', 'callable', 'enabled', 'timeout']
        for attr in attrs:
            self.assertEqual(getattr(jobs[0], attr), getattr(db_job, attr))

    def test_import__should_schedule_job_yaml(self):
        jobs = list()
        jobs.append(task_factory(ScheduledTask, enabled=True, instance_only=True))
        jobs.append(task_factory(RepeatableTask, enabled=True, instance_only=True))
        res = yaml.dump([j.to_dict() for j in jobs], default_flow_style=False)
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        call_command('import', filename=self.tmpfile.name, format='yaml')
        # assert
        self.assertEqual(1, ScheduledTask.objects.count())
        db_job = ScheduledTask.objects.first()
        attrs = ['name', 'queue', 'callable', 'enabled', 'timeout']
        for attr in attrs:
            self.assertEqual(getattr(jobs[0], attr), getattr(db_job, attr))

    def test_import__should_schedule_job_yaml_without_yaml_lib(self):
        jobs = list()
        jobs.append(task_factory(ScheduledTask, enabled=True, instance_only=True))
        jobs.append(task_factory(RepeatableTask, enabled=True, instance_only=True))
        res = yaml.dump([j.to_dict() for j in jobs], default_flow_style=False)
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        with mock.patch.dict('sys.modules', {'yaml': None}):
            with self.assertRaises(SystemExit) as cm:
                call_command('import', filename=self.tmpfile.name, format='yaml')
            self.assertEqual(cm.exception.code, 1)

    def test_import__should_schedule_job_reset(self):
        jobs = list()
        task_factory(ScheduledTask, enabled=True)
        task_factory(ScheduledTask, enabled=True)
        jobs.append(task_factory(ScheduledTask, enabled=True))
        jobs.append(task_factory(RepeatableTask, enabled=True, instance_only=True))
        res = json.dumps([j.to_dict() for j in jobs])
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        call_command('import', filename=self.tmpfile.name, reset=True, )
        # assert
        self.assertEqual(1, ScheduledTask.objects.count())
        db_job = ScheduledTask.objects.first()
        attrs = ['name', 'queue', 'callable', 'enabled', 'timeout']
        for attr in attrs:
            self.assertEqual(getattr(jobs[0], attr), getattr(db_job, attr))
        self.assertEqual(1, RepeatableTask.objects.count())
        db_job = RepeatableTask.objects.first()
        attrs = ['name', 'queue', 'callable', 'enabled', 'timeout']
        for attr in attrs:
            self.assertEqual(getattr(jobs[1], attr), getattr(db_job, attr))

    def test_import__should_schedule_job_update_existing(self):
        jobs = list()
        task_factory(ScheduledTask, enabled=True)
        jobs.append(task_factory(ScheduledTask, enabled=True))
        res = json.dumps([j.to_dict() for j in jobs])
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        call_command('import', filename=self.tmpfile.name, update=True, )
        # assert
        self.assertEqual(2, ScheduledTask.objects.count())
        db_job = ScheduledTask.objects.get(name=jobs[0].name)
        self.assertNotEqual(jobs[0].id, db_job.id)
        attrs = ['name', 'queue', 'callable', 'enabled', 'timeout']
        for attr in attrs:
            self.assertEqual(getattr(jobs[0], attr), getattr(db_job, attr))

    def test_import__should_schedule_job_without_update_existing(self):
        jobs = list()
        task_factory(ScheduledTask, enabled=True)
        jobs.append(task_factory(ScheduledTask, enabled=True))
        res = json.dumps([j.to_dict() for j in jobs])
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        call_command('import', filename=self.tmpfile.name, )
        # assert
        self.assertEqual(2, ScheduledTask.objects.count())
        db_job = ScheduledTask.objects.get(name=jobs[0].name)
        attrs = ['id', 'name', 'queue', 'callable', 'enabled', 'timeout']
        for attr in attrs:
            self.assertEqual(getattr(jobs[0], attr), getattr(db_job, attr))



from django.urls import path

from . import views

urlpatterns = [
    path('queues/', views.stats, name='queues_home'),
    path('queues/stats.json', views.stats_json, name='queues_home_json'),
    path('queues/<str:queue_name>/workers/', views.queue_workers, name='queue_workers'),
    path('queues/<str:queue_name>/<str:registry_name>/jobs', views.jobs_view, name='queue_registry_jobs'),
    path('queues/<str:queue_name>/<str:registry_name>/empty/', views.clear_queue_registry, name='queue_clear'),
    path('queues/<str:queue_name>/<str:registry_name>/requeue-all/', views.requeue_all, name='queue_requeue_all'),
    path('queues/<str:queue_name>/confirm-action/', views.confirm_action, name='queue_confirm_action'),
    path('queues/<str:queue_name>/actions/', views.actions, name='queue_actions'),
]

urlpatterns += [
    path('workers/', views.workers, name='workers_home'),
    path('workers/<str:name>/', views.worker_details, name='worker_details'),
    path('jobs/<str:job_id>/', views.job_detail, name='job_details'),
    path('jobs/<str:job_id>/<str:action>/', views.job_action, name='queue_job_action'),
]



import click
from django.core.management.base import BaseCommand

from scheduler.queues import get_queue


class Command(BaseCommand):
    """
    Queues the function given with the first argument with the
    parameters given with the rest of the argument list.
    """
    help = __doc__
    args = '<function arg arg ...>'

    def add_arguments(self, parser):
        parser.add_argument(
            '--queue', '-q', dest='queue', default='default',
            help='Specify the queue [default]')
        parser.add_argument(
            '--timeout', '-t', type=int, dest='timeout',
            help='A timeout in seconds')
        parser.add_argument(
            '--result-ttl', '-r', type=int, dest='result_ttl',
            help='Time to store job results in seconds')
        parser.add_argument('callable', help='Method to call', )
        parser.add_argument('args', nargs='*', help='Args for callable')

    def handle(self, **options):
        verbosity = int(options.get('verbosity', 1))
        timeout = options.get('timeout')
        result_ttl = options.get('result_ttl')
        queue = get_queue(options.get('queue'))
        func = options.get('callable')
        args = options.get('args')
        job = queue.enqueue_call(func, args=args, timeout=timeout, result_ttl=result_ttl)
        if verbosity:
            click.echo(f'Job {job.id} created')



# Generated by Django 4.1.4 on 2022-12-18 18:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('scheduler', '0003_auto_20220329_2107'),
    ]

    operations = [
        migrations.AddField(
            model_name='cronjob',
            name='at_front',
            field=models.BooleanField(default=False, verbose_name='At front'),
        ),
        migrations.AddField(
            model_name='repeatablejob',
            name='at_front',
            field=models.BooleanField(default=False, verbose_name='At front'),
        ),
        migrations.AddField(
            model_name='scheduledjob',
            name='at_front',
            field=models.BooleanField(default=False, verbose_name='At front'),
        ),
    ]



import logging
import os
import sys

import click
from django.core.management.base import BaseCommand
from django.db import connections
from redis.exceptions import ConnectionError
from rq.logutils import setup_loghandlers

from scheduler.tools import create_worker

VERBOSITY_TO_LOG_LEVEL = {
    0: logging.CRITICAL,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}


def reset_db_connections():
    for c in connections.all():
        c.close()


class Command(BaseCommand):
    """
    Runs RQ workers on specified queues. Note that all queues passed into a
    single rqworker command must share the same connection.

    Example usage:
    python manage.py rqworker high medium low
    """

    args = '<queue queue ...>'

    def add_arguments(self, parser):
        parser.add_argument('--pid', action='store', dest='pidfile',
                            default=None, help='file to write the worker`s pid into')
        parser.add_argument('--burst', action='store_true', dest='burst',
                            default=False, help='Run worker in burst mode')
        parser.add_argument('--name', action='store', dest='name',
                            default=None, help='Name of the worker')
        parser.add_argument('--worker-ttl', action='store', type=int, dest='worker_ttl', default=420,
                            help='Default worker timeout to be used')
        parser.add_argument('--max-jobs', action='store', default=None, dest='max_jobs', type=int,
                            help='Maximum number of jobs to execute before terminating worker')
        parser.add_argument('--fork-job-execution', action='store', default=True, dest='fork_job_execution', type=bool,
                            help='Fork job execution to another process')
        parser.add_argument(
            'queues', nargs='*', type=str,
            help='The queues to work on, separated by space, all queues should be using the same redis')

    def handle(self, **options):
        queues = options.get('queues', [])
        if not queues:
            queues = ['default', ]
        click.echo(f'Starting worker for queues {queues}')
        pidfile = options.get('pidfile')
        if pidfile:
            with open(os.path.expanduser(pidfile), "w") as fp:
                fp.write(str(os.getpid()))

        # Verbosity is defined by default in BaseCommand for all commands
        verbosity = options.get('verbosity', 1)
        log_level = VERBOSITY_TO_LOG_LEVEL.get(verbosity, logging.INFO)
        setup_loghandlers(log_level)

        try:
            # Instantiate a worker
            w = create_worker(
                *queues,
                name=options['name'],
                default_worker_ttl=options['worker_ttl'],
                fork_job_execution=options['fork_job_execution'], )

            # Close any opened DB connection before any fork
            reset_db_connections()

            w.work(burst=options.get('burst', False),
                   logging_level=log_level,
                   max_jobs=options['max_jobs'], )
        except ConnectionError as e:
            click.echo(str(e), err=True)
            sys.exit(1)



import logging

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

logger = logging.getLogger(__package__)

QUEUES = dict()
SCHEDULER_CONFIG = dict()


def _token_validation(token: str) -> bool:
    return False


def conf_settings():
    global QUEUES
    global SCHEDULER_CONFIG

    QUEUES = getattr(settings, 'SCHEDULER_QUEUES', None)
    if QUEUES is None:
        logger.warning('Configuration using RQ_QUEUES is deprecated. Use SCHEDULER_QUEUES instead')
        QUEUES = getattr(settings, 'RQ_QUEUES', None)
    if QUEUES is None:
        raise ImproperlyConfigured("You have to define SCHEDULER_QUEUES in settings.py")

    SCHEDULER_CONFIG = {
        'EXECUTIONS_IN_PAGE': 20,
        'DEFAULT_RESULT_TTL': 600,  # 10 minutes
        'DEFAULT_TIMEOUT': 300,  # 5 minutes
        'SCHEDULER_INTERVAL': 10,  # 10 seconds
        'FAKEREDIS': False,  # For testing purposes
        'TOKEN_VALIDATION_METHOD': _token_validation,  # Access stats from another application using API tokens
    }
    user_settings = getattr(settings, 'SCHEDULER_CONFIG', {})
    SCHEDULER_CONFIG.update(user_settings)


conf_settings()


def get_config(key: str, default=None):
    return SCHEDULER_CONFIG.get(key, None)



from time import sleep

from scheduler.queues import get_queue

_counter = 0


def arg_callable():
    global _counter
    _counter += 1
    return _counter


def test_args_kwargs(*args, **kwargs):
    func = "test_args_kwargs({})"
    args_list = [repr(arg) for arg in args]
    kwargs_list = [f'{k}={v}' for (k, v) in kwargs.items()]
    return func.format(', '.join(args_list + kwargs_list))


def long_job():
    sleep(10)


test_non_callable = 'I am a teapot'


def failing_job():
    raise ValueError


def test_job():
    return 1 + 1


def enqueue_jobs():
    queue = get_queue()
    for i in range(20):
        queue.enqueue(test_job, job_id=f"job_{i}", args=())



# Generated by Django 4.2.5 on 2023-10-08 16:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('scheduler', '0014_alter_cronjob_created_alter_cronjob_modified_and_more'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='CronJob',
            new_name='CronTask',
        ),
        migrations.RenameModel(
            old_name='RepeatableJob',
            new_name='RepeatableTask',
        ),
        migrations.RenameModel(
            old_name='ScheduledJob',
            new_name='ScheduledTask',
        ),
    ]



import sys

import click
from django.apps import apps
from django.core.management.base import BaseCommand

from scheduler.tools import MODEL_NAMES


class Command(BaseCommand):
    """
    Export all scheduled jobs
    """
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            '-o', '--output',
            action='store',
            choices=['json', 'yaml'],
            default='json',
            dest='format',
            help='format of output',
        )

        parser.add_argument(
            '-e', '--enabled',
            action='store_true',
            dest='enabled',
            help='Export only enabled jobs',
        )
        parser.add_argument(
            '-f', '--filename',
            action='store',
            dest='filename',
            help='File name to load (otherwise writes to standard output)',
        )

    def handle(self, *args, **options):
        file = open(options.get('filename'), 'w') if options.get("filename") else sys.stdout
        res = list()
        for model_name in MODEL_NAMES:
            model = apps.get_model(app_label='scheduler', model_name=model_name)
            jobs = model.objects.all()
            if options.get('enabled'):
                jobs = jobs.filter(enabled=True)
            for job in jobs:
                res.append(job.to_dict())

        if options.get("format") == 'json':
            import json
            click.echo(json.dumps(res, indent=2), file=file)
            return

        if options.get("format") == 'yaml':
            try:
                import yaml
            except ImportError:
                click.echo("Aborting. LibYAML is not installed.")
                exit(1)
            # Disable YAML alias
            yaml.Dumper.ignore_aliases = lambda *x: True
            click.echo(yaml.dump(res, default_flow_style=False), file=file)
            return



import uuid
from datetime import datetime
from unittest.mock import patch, PropertyMock

from django.contrib.auth.models import User
from django.test import TestCase
from django.test.client import Client
from django.urls import reverse

from scheduler.queues import get_queue
from scheduler.tools import create_worker
from . import test_settings  # noqa
from .jobs import failing_job, long_job, test_job
from .testtools import assert_message_in_response, task_factory, _get_job_from_scheduled_registry
from ..models import ScheduledTask
from ..rq_classes import JobExecution, ExecutionStatus


class BaseTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_superuser('user', password='pass')
        self.client = Client()
        self.client.login(username=self.user.username, password='pass')
        get_queue('django_tasks_scheduler_test').connection.flushall()


class SingleJobActionViewsTest(BaseTestCase):

    def test_single_job_action_unknown_job(self):
        res = self.client.get(reverse('queue_job_action', args=['unknown', 'cancel']), follow=True)
        self.assertEqual(400, res.status_code)

    def test_single_job_action_unknown_action(self):
        queue = get_queue('default')
        job = queue.enqueue(failing_job)
        worker = create_worker('default')
        worker.work(burst=True)
        job.refresh()
        self.assertTrue(job.is_failed)
        res = self.client.get(reverse('queue_job_action', args=[job.id, 'unknown']), follow=True)
        self.assertEqual(404, res.status_code)

    def test_single_job_action_requeue_job(self):
        queue = get_queue('default')
        job = queue.enqueue(failing_job)
        worker = create_worker('default')
        worker.work(burst=True)
        job.refresh()
        self.assertTrue(job.is_failed)
        res = self.client.get(reverse('queue_job_action', args=[job.id, 'requeue']), follow=True)
        self.assertEqual(200, res.status_code)
        self.client.post(reverse('queue_job_action', args=[job.id, 'requeue']), {'requeue': 'Requeue'}, follow=True)
        self.assertIn(job, queue.jobs)
        job.delete()

    def test_single_job_action_delete_job(self):
        queue = get_queue('django_tasks_scheduler_test')
        job = queue.enqueue(test_job)
        res = self.client.get(reverse('queue_job_action', args=[job.id, 'delete']), follow=True)
        self.assertEqual(200, res.status_code)
        self.client.post(reverse('queue_job_action', args=[job.id, 'delete']), {'post': 'yes'}, follow=True)
        self.assertFalse(JobExecution.exists(job.id, connection=queue.connection))
        self.assertNotIn(job.id, queue.get_job_ids())

    def test_single_job_action_cancel_job(self):
        queue = get_queue('django_tasks_scheduler_test')
        job = queue.enqueue(long_job)
        res = self.client.get(reverse('queue_job_action', args=[job.id, 'cancel']), follow=True)
        self.assertEqual(200, res.status_code)
        res = self.client.post(reverse('queue_job_action', args=[job.id, 'cancel']), {'post': 'yes'}, follow=True)
        self.assertEqual(200, res.status_code)
        tmp = JobExecution.fetch(job.id, connection=queue.connection)
        self.assertTrue(tmp.is_canceled)
        self.assertNotIn(job.id, queue.get_job_ids())

    def test_single_job_action_cancel_job_that_is_already_cancelled(self):
        queue = get_queue('django_tasks_scheduler_test')
        job = queue.enqueue(long_job)
        res = self.client.post(reverse('queue_job_action', args=[job.id, 'cancel']), {'post': 'yes'}, follow=True)
        self.assertEqual(200, res.status_code)
        tmp = JobExecution.fetch(job.id, connection=queue.connection)
        self.assertTrue(tmp.is_canceled)
        self.assertNotIn(job.id, queue.get_job_ids())
        res = self.client.post(reverse('queue_job_action', args=[job.id, 'cancel']), {'post': 'yes'}, follow=True)
        self.assertEqual(200, res.status_code)
        assert_message_in_response(res, f'Could not perform action: Cannot cancel already canceled job: {job.id}')

    def test_single_job_action_enqueue_job(self):
        queue = get_queue('django_tasks_scheduler_test')
        job_list = []
        # enqueue some jobs that depends on other
        previous_job = None
        for _ in range(0, 3):
            job = queue.enqueue(test_job, depends_on=previous_job)
            job_list.append(job)
            previous_job = job

        # This job is deferred

        self.assertEqual(job_list[-1].get_status(), ExecutionStatus.DEFERRED)
        self.assertIsNone(job_list[-1].enqueued_at)

        # Try to force enqueue last job should do nothing
        res = self.client.get(reverse('queue_job_action', args=[job_list[-1].id, 'enqueue']), follow=True)
        self.assertEqual(200, res.status_code)
        res = self.client.post(reverse('queue_job_action', args=[job_list[-1].id, 'enqueue']), follow=True)

        # Check that job is still deferred because it has dependencies (rq 1.14 change)
        self.assertEqual(200, res.status_code)
        tmp = queue.fetch_job(job_list[-1].id)
        self.assertEqual(tmp.get_status(), ExecutionStatus.QUEUED)
        self.assertIsNotNone(tmp.enqueued_at)


class JobListActionViewsTest(BaseTestCase):
    def test_job_list_action_delete_jobs__with_bad_next_url(self):
        queue = get_queue('django_tasks_scheduler_test')

        # enqueue some jobs
        job_ids = []
        for _ in range(0, 3):
            job = queue.enqueue(test_job)
            job_ids.append(job.id)

        # remove those jobs using view
        res = self.client.post(
            reverse('queue_actions', args=[queue.name, ]), {
                'action': 'delete', 'job_ids': job_ids,
                'next_url': 'bad_url',
            }, follow=True)
        assert_message_in_response(res, 'Bad followup URL')
        # check if jobs are removed
        self.assertEqual(200, res.status_code)
        for job_id in job_ids:
            self.assertFalse(JobExecution.exists(job_id, connection=queue.connection))
            self.assertNotIn(job_id, queue.job_ids)

    def test_job_list_action_delete_jobs(self):
        queue = get_queue('django_tasks_scheduler_test')

        # enqueue some jobs
        job_ids = []
        for _ in range(0, 3):
            job = queue.enqueue(test_job)
            job_ids.append(job.id)

        # remove those jobs using view
        res = self.client.post(
            reverse('queue_actions', args=[queue.name, ]), {'action': 'delete', 'job_ids': job_ids}, follow=True)

        # check if jobs are removed
        self.assertEqual(200, res.status_code)
        for job_id in job_ids:
            self.assertFalse(JobExecution.exists(job_id, connection=queue.connection))
            self.assertNotIn(job_id, queue.job_ids)

    def test_job_list_action_requeue_jobs(self):
        queue = get_queue('django_tasks_scheduler_test')
        queue_name = 'django_tasks_scheduler_test'

        # enqueue some jobs that will fail
        jobs = []
        job_ids = []
        for _ in range(0, 3):
            job = queue.enqueue(failing_job)
            jobs.append(job)
            job_ids.append(job.id)

        # do those jobs = fail them
        worker = create_worker('django_tasks_scheduler_test')
        worker.work(burst=True)

        # check if all jobs are really failed
        for job in jobs:
            self.assertTrue(job.is_failed)

        # re-nqueue failed jobs from failed queue
        self.client.post(reverse('queue_actions', args=[queue_name]), {'action': 'requeue', 'job_ids': job_ids})

        # check if we requeue all failed jobs
        for job in jobs:
            self.assertFalse(job.is_failed)

    def test_job_list_action_stop_jobs(self):
        queue_name = 'django_tasks_scheduler_test'
        queue = get_queue(queue_name)

        # Enqueue some jobs
        job_ids = []
        worker = create_worker('django_tasks_scheduler_test')
        for _ in range(3):
            job = queue.enqueue(test_job)
            job_ids.append(job.id)
            worker.prepare_job_execution(job)

        # Check if the jobs are started
        for job_id in job_ids:
            job = JobExecution.fetch(job_id, connection=queue.connection)
            self.assertEqual(job.get_status(), ExecutionStatus.STARTED)

        # Stop those jobs using the view
        started_job_registry = queue.started_job_registry
        self.assertEqual(len(started_job_registry), len(job_ids))
        self.client.post(reverse('queue_actions', args=[queue_name]), {'action': 'stop', 'job_ids': job_ids})
        self.assertEqual(len(started_job_registry), 0)

        canceled_job_registry = queue.canceled_job_registry
        self.assertEqual(len(canceled_job_registry), len(job_ids))

        for job_id in job_ids:
            self.assertIn(job_id, canceled_job_registry)


class QueueRegistryJobsViewTest(BaseTestCase):
    def test_queue_jobs_unknown_registry(self):
        queue_name = 'default'
        res = self.client.get(reverse('queue_registry_jobs', args=[queue_name, 'unknown']), follow=True)
        self.assertEqual(404, res.status_code)

    def test_queue_jobs_unknown_queue(self):
        res = self.client.get(reverse('queue_registry_jobs', args=['UNKNOWN', 'queued']))
        self.assertEqual(404, res.status_code)

    def test_queued_jobs(self):
        """Jobs in queue are displayed properly"""
        queue = get_queue('default')
        job = queue.enqueue(test_job)
        queue_name = 'default'
        res = self.client.get(reverse('queue_registry_jobs', args=[queue_name, 'queued']))
        self.assertEqual(res.context['jobs'], [job])

    def test_finished_jobs(self):
        """Ensure that finished jobs page works properly."""
        queue = get_queue('django_tasks_scheduler_test')
        queue_name = 'django_tasks_scheduler_test'

        job = queue.enqueue(test_job)
        registry = queue.finished_job_registry
        registry.add(job, 2)
        res = self.client.get(reverse('queue_registry_jobs', args=[queue_name, 'finished']))
        self.assertEqual(res.context['jobs'], [job])

    def test_failed_jobs(self):
        """Ensure that failed jobs page works properly."""
        queue = get_queue('django_tasks_scheduler_test')
        queue_name = 'django_tasks_scheduler_test'

        # Test that page doesn't fail when FailedJobRegistry is empty
        res = self.client.get(reverse('queue_registry_jobs', args=[queue_name, 'failed']))
        self.assertEqual(res.status_code, 200)

        job = queue.enqueue(test_job)
        registry = queue.failed_job_registry
        registry.add(job, 2)
        res = self.client.get(reverse('queue_registry_jobs', args=[queue_name, 'failed']))
        self.assertEqual(res.context['jobs'], [job])

    def test_scheduled_jobs(self):
        """Ensure that scheduled jobs page works properly."""
        queue = get_queue('django_tasks_scheduler_test')
        queue_name = 'django_tasks_scheduler_test'

        # Test that page doesn't fail when ScheduledJobRegistry is empty
        res = self.client.get(reverse('queue_registry_jobs', args=[queue_name, 'scheduled']))
        self.assertEqual(res.status_code, 200)

        job = queue.enqueue_at(datetime.now(), test_job)
        res = self.client.get(reverse('queue_registry_jobs', args=[queue_name, 'scheduled']))
        self.assertEqual(res.context['jobs'], [job])

    def test_scheduled_jobs_registry_removal(self):
        """Ensure that non-existing job is being deleted from registry by view"""
        queue = get_queue('django_tasks_scheduler_test')
        queue_name = 'django_tasks_scheduler_test'

        registry = queue.scheduled_job_registry
        job = queue.enqueue_at(datetime.now(), test_job)
        self.assertEqual(len(registry), 1)

        queue.connection.delete(job.key)
        res = self.client.get(reverse('queue_registry_jobs', args=[queue_name, 'scheduled']))
        self.assertEqual(res.context['jobs'], [])

        self.assertEqual(len(registry), 0)

    def test_started_jobs(self):
        """Ensure that active jobs page works properly."""
        queue = get_queue('django_tasks_scheduler_test')
        queue_name = 'django_tasks_scheduler_test'

        job = queue.enqueue(test_job)
        registry = queue.started_job_registry
        registry.add(job, 2)
        res = self.client.get(reverse('queue_registry_jobs', args=[queue_name, 'started']))
        self.assertEqual(res.context['jobs'], [job])

    def test_deferred_jobs(self):
        """Ensure that active jobs page works properly."""
        queue = get_queue('django_tasks_scheduler_test')
        queue_name = 'django_tasks_scheduler_test'

        job = queue.enqueue(test_job)
        registry = queue.deferred_job_registry
        registry.add(job, 2)
        res = self.client.get(reverse('queue_registry_jobs', args=[queue_name, 'deferred']))
        self.assertEqual(res.context['jobs'], [job])


class ViewTest(BaseTestCase):

    def test_job_details(self):
        """Job data is displayed properly"""
        queue = get_queue('default')
        job = queue.enqueue(test_job)

        url = reverse('job_details', args=[job.id, ])
        res = self.client.get(url)
        self.assertIn('job', res.context)
        self.assertEqual(res.context['job'], job)

        # This page shouldn't fail when job.data is corrupt
        queue.connection.hset(job.key, 'data', 'non-pickleable data')
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)
        self.assertIn('DeserializationError', res.content.decode())

        # Bad job-id should return 404
        url = reverse('job_details', args=['bad_job_id', ])
        res = self.client.get(url)
        self.assertEqual(400, res.status_code)

    def test_scheduled_job_details(self):
        """Job data is displayed properly"""
        scheduled_job = task_factory(ScheduledTask, enabled=True)
        job = _get_job_from_scheduled_registry(scheduled_job)

        url = reverse('job_details', args=[job.id, ])
        res = self.client.get(url, follow=True)
        self.assertIn('job', res.context)
        self.assertEqual(res.context['job'], job)

    def test_job_details_on_deleted_dependency(self):
        """Page doesn't crash even if job.dependency has been deleted"""
        queue = get_queue('default')

        job = queue.enqueue(test_job)
        second_job = queue.enqueue(test_job, depends_on=job)
        job.delete()
        url = reverse('job_details', args=[second_job.id])
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)
        self.assertIn(second_job._dependency_id, res.content.decode())

    def test_requeue_all(self):
        """
        Ensure that re-queuing all failed job work properly
        """
        queue = get_queue('default')
        queue_name = 'default'
        queue.enqueue(failing_job)
        queue.enqueue(failing_job)
        worker = create_worker('default')
        worker.work(burst=True)

        res = self.client.get(reverse('queue_requeue_all', args=[queue_name, 'failed']))
        self.assertEqual(res.context['total_jobs'], 2)
        # After requeue_all is called, jobs are enqueued
        res = self.client.post(reverse('queue_requeue_all', args=[queue_name, 'failed']))
        self.assertEqual(len(queue), 2)

    def test_requeue_all_if_deleted_job(self):
        """
        Ensure that re-queuing all failed job work properly
        """
        queue = get_queue('default')
        queue_name = 'default'
        job = queue.enqueue(failing_job)
        queue.enqueue(failing_job)
        worker = create_worker('default')
        worker.work(burst=True)

        res = self.client.get(reverse('queue_requeue_all', args=[queue_name, 'failed']))
        self.assertEqual(res.context['total_jobs'], 2)
        job.delete()

        # After requeue_all is called, jobs are enqueued
        res = self.client.post(reverse('queue_requeue_all', args=[queue_name, 'failed']))
        self.assertEqual(len(queue), 1)

    def test_clear_queue_unknown_registry(self):
        queue_name = 'django_tasks_scheduler_test'
        res = self.client.post(reverse('queue_clear', args=[queue_name, 'unknown']), {'post': 'yes'})
        self.assertEqual(404, res.status_code)

    def test_clear_queue_enqueued(self):
        queue = get_queue('django_tasks_scheduler_test')
        job = queue.enqueue(test_job)
        self.client.post(reverse('queue_clear', args=[queue.name, 'queued']), {'post': 'yes'})
        self.assertFalse(JobExecution.exists(job.id, connection=queue.connection))
        self.assertNotIn(job.id, queue.job_ids)

    def test_clear_queue_scheduled(self):
        queue = get_queue('django_tasks_scheduler_test')
        job = queue.enqueue_at(datetime.now(), test_job)

        res = self.client.get(reverse('queue_clear', args=[queue.name, 'scheduled']), follow=True)
        self.assertEqual(200, res.status_code)
        self.assertEqual(res.context['jobs'], [job, ])

        res = self.client.post(reverse('queue_clear', args=[queue.name, 'scheduled']), {'post': 'yes'}, follow=True)
        assert_message_in_response(res, f'You have successfully cleared the scheduled jobs in queue {queue.name}')
        self.assertEqual(200, res.status_code)
        self.assertFalse(JobExecution.exists(job.id, connection=queue.connection))
        self.assertNotIn(job.id, queue.job_ids)

    def test_workers_home(self):
        res = self.client.get(reverse('workers_home'))
        prev_workers = res.context['workers']
        worker1 = create_worker('django_tasks_scheduler_test')
        worker1.register_birth()
        worker2 = create_worker('test3')
        worker2.register_birth()

        res = self.client.get(reverse('workers_home'))
        self.assertEqual(res.context['workers'], prev_workers + [worker1, worker2])

    def test_queue_workers(self):
        """Worker index page should show workers for a specific queue"""
        queue_name = 'django_tasks_scheduler_test'

        worker1 = create_worker('django_tasks_scheduler_test')
        worker1.register_birth()
        worker2 = create_worker('test3')
        worker2.register_birth()

        res = self.client.get(reverse('queue_workers', args=[queue_name]))
        self.assertEqual(res.context['workers'], [worker1])

    def test_worker_details(self):
        """Worker index page should show workers for a specific queue"""

        worker = create_worker('django_tasks_scheduler_test', name=uuid.uuid4().hex)
        worker.register_birth()

        url = reverse('worker_details', args=[worker.name, ])
        res = self.client.get(url)
        self.assertEqual(res.context['worker'], worker)

    def test_worker_details__non_existing_worker(self):
        """Worker index page should show workers for a specific queue"""

        worker = create_worker('django_tasks_scheduler_test', name='WORKER')
        worker.register_birth()

        res = self.client.get(reverse('worker_details', args=['bad-worker-name']))
        self.assertEqual(404, res.status_code)

    def test_statistics_json_view(self):
        # Override testing SCHEDULER_QUEUES
        queues = {
            'default': {
                'DB': 0,
                'HOST': 'localhost',
                'PORT': 6379,
            }
        }
        with patch('scheduler.settings.QUEUES', new_callable=PropertyMock(return_value=queues)):
            res = self.client.get(reverse('queues_home'))
            self.assertEqual(res.status_code, 200)

            res = self.client.get(reverse('queues_home_json'))
            self.assertEqual(res.status_code, 200)

            # Not staff => return 404
            self.user.is_staff = False
            self.user.save()

            res = self.client.get(reverse('queues_home'))
            self.assertEqual(res.status_code, 302)

            # 404 code for stats
            res = self.client.get(reverse('queues_home_json'))
            self.assertEqual(res.status_code, 404)

    @staticmethod
    def token_validation(token: str) -> bool:
        return token == 'valid'

    # @patch('scheduler.views.SCHEDULER_CONFIG')
    # def test_statistics_json_view_token(self, configuration):
    #     configuration.get.return_value = ViewTest.token_validation
    #     self.user.is_staff = False
    #     self.user.save()
    #     res = self.client.get(reverse('queues_home_json'), headers={'Authorization': 'valid'})
    #     self.assertEqual(res.status_code, 200)
    #
    #     res = self.client.get(reverse('queues_home_json'), headers={'Authorization': 'invalid'})
    #     self.assertEqual(res.status_code, 404)
